#include "tmesh.h"
#include "v3.h"

#include <fstream>
#include <iostream>

#include <float.h>


TMesh::TMesh(V3 center, float dims[2]) {

  texture = 0;
  verticesN = 4;
  trianglesN = 2;
  vertices = new V3[verticesN];
  projVertices = new V3[verticesN];
  colors = 0;
  tcs = 0;
  connectivity = new unsigned int[2*3];
  vertices[0] = center+V3(-dims[0]/2.0f, -dims[1]/2.0f, 0.0f);
  vertices[1] = center+V3(+dims[0]/2.0f, -dims[1]/2.0f, 0.0f);
  vertices[2] = center+V3(+dims[0]/2.0f, +dims[1]/2.0f, 0.0f);
  vertices[3] = center+V3(-dims[0]/2.0f, +dims[1]/2.0f, 0.0f);

  connectivity[0] = 0;
  connectivity[1] = 1;
  connectivity[2] = 2;
  connectivity[3] = 2;
  connectivity[4] = 3;
  connectivity[5] = 0;

}

void TMesh::SetColors(V3 *cols) {

  if (!colors)
    colors = new V3[verticesN];
  for (int i = 0; i < verticesN; i++)
    colors[i] = cols[i];

}

TMesh::TMesh() : vertices(0), verticesN(0), colors(0), connectivity(0),
  trianglesN(0), tcs(0), texture(0), projVertices(0) {}

void TMesh::Load(char *fname) {

  ifstream ifs(fname, ios::binary);
  if (ifs.fail()) {
    cerr << "INFO: cannot open file: " << fname << endl;
    return;
  }

  ifs.read((char*)&verticesN, sizeof(int));
  char yn;
  ifs.read(&yn, 1); // always xyz
  if (yn != 'y') {
    cerr << "INTERNAL ERROR: there should always be vertex xyz data" << endl;
    return;
  }
  if (vertices)
    delete vertices;
  vertices = new V3[verticesN];
  projVertices = new V3[verticesN];

  ifs.read(&yn, 1); // colors 3 floats
  if (colors)
    delete colors;
  colors = 0;
  if (yn == 'y') {
    colors = new V3[verticesN];
  }

  ifs.read(&yn, 1); // normals 3 floats
  V3 *normals = 0; // don't have normals for now
  if (normals)
    delete normals;
  normals = 0;
  if (yn == 'y') {
    normals = new V3[verticesN];
  }

  ifs.read(&yn, 1); // texture coordinates 2 floats
  float *tcs = 0; // don't have texture coordinates for now
  if (tcs)
    delete tcs;
  tcs = 0;
  if (yn == 'y') {
    tcs = new float[verticesN*2];
  }

  ifs.read((char*)vertices, verticesN*3*sizeof(float)); // load vertices

  if (colors) {
    ifs.read((char*)colors, verticesN*3*sizeof(float)); // load colors
  }

  if (normals)
    ifs.read((char*)normals, verticesN*3*sizeof(float)); // load normals

  if (tcs)
    ifs.read((char*)tcs, verticesN*2*sizeof(float)); // load texture coordinates

  ifs.read((char*)&trianglesN, sizeof(int));
  if (connectivity)
    delete connectivity;
  connectivity = new unsigned int[trianglesN*3];
  ifs.read((char*)connectivity, trianglesN*3*sizeof(unsigned int)); // read tiangles

  ifs.close();

  cerr << "INFO: loaded " << verticesN << " verts, " << trianglesN << " tris from " << endl << "      " << fname << endl;
  cerr << "      xyz " << ((colors) ? "rgb " : "") << ((normals) ? "nxnynz " : "") << ((tcs) ? "tcstct " : "") << endl;

  if (normals)
    delete normals;
  if (tcs)
    delete tcs;

}

void TMesh::RenderPoints(PPC *ppc, FrameBuffer *fb) {

  for (int i = 0; i < verticesN; i++) {
    if (projVertices[i][0] == FLT_MAX)
      continue;
    fb->SetIfCloser(projVertices[i], colors[i]);
  }

}

void TMesh::Project(PPC *ppc) {

  for (int i = 0; i < verticesN; i++) {
    ppc->Project(vertices[i], projVertices[i]);
  }

}

void TMesh::Render(PPC *ppc, FrameBuffer *fb,
                   PPC *projPPC, FrameBuffer *projFB) {

  Project(ppc);

  for (int tri = 0; tri < trianglesN; tri++) {
    unsigned int inds[3] = {connectivity[3*tri+0], 
      connectivity[3*tri+1], connectivity[3*tri+2]};
    V3 pvs[3] = {projVertices[inds[0]], projVertices[inds[1]], projVertices[inds[2]]};

    // return if any vertex behind head
    if (pvs[0][0] == FLT_MAX)
      continue;
    if (pvs[1][0] == FLT_MAX)
      continue;
    if (pvs[2][0] == FLT_MAX)
      continue;

    // compute bounding box
    BBox bbox;
    bbox.AddPoint(pvs[0]);
    bbox.AddPoint(pvs[1]);
    bbox.AddPoint(pvs[2]);
    BBox imageFrame;
    float border = 0.0f;
    imageFrame.AddPoint(V3(0.0f+border, 0.0f+border, -FLT_MAX)); 
    imageFrame.AddPoint(V3((float)fb->w-1.0f-border, (float)fb->h-1.0f-border, FLT_MAX));
    if (!bbox.Clip(imageFrame))
      continue;
    int left = (int) (bbox.corners[0][0]+0.5f);
    int right = (int) (bbox.corners[1][0]-0.5f);
    int top = (int) (bbox.corners[0][1]+0.5f);
    int bottom = (int) (bbox.corners[1][1]-0.5f);

    // rasterization setup
    M3x3 barycentricInverse = ComputeBarycentricInverse(pvs);
    M3x3 ssiM = ComputeSSI(pvs);

    M3x3 qM = ComputeMMI(inds, ppc);
    V3 msiDenABC(qM[0][0]+qM[1][0]+qM[2][0],
      qM[0][1]+qM[1][1]+qM[2][1], qM[0][2]+qM[1][2]+qM[2][2]);

    V3 zABC = ssiM*V3(pvs[0][2], pvs[1][2], pvs[2][2]);
    M3x3 cols;
    cols[0] = colors[inds[0]];
    cols[1] = colors[inds[1]];
    cols[2] = colors[inds[2]];
    M3x3 colABC = (ssiM*cols);
    colABC.Transpose();

    V3 redABC = GetNumABC(qM, cols.GetColumn(0));
    V3 greenABC = GetNumABC(qM, cols.GetColumn(1));
    V3 blueABC = GetNumABC(qM, cols.GetColumn(2));

    V3 texABCs[2];
    V3 texABCsm[2];
    if (texture) {
      V3 texSs(tcs[inds[0]*2+0], tcs[inds[1]*2+0], tcs[inds[2]*2+0]);
      V3 texTs(tcs[inds[0]*2+1], tcs[inds[1]*2+1], tcs[inds[2]*2+1]);
      texABCs[0] = ssiM * texSs;
      texABCs[1] = ssiM * texTs;
      texABCsm[0] = GetNumABC(qM, texSs);
      texABCsm[1] = GetNumABC(qM, texTs);
    }

    for (int v = top; v <= bottom; v++) {
      for (int u = left; u <= right; u++) {
        V3 pixv(.5f+(float)u, .5f+(float)v, 1.0f);
        V3 baryCoords = barycentricInverse*pixv;
        if (baryCoords[0] < 0.0f || baryCoords[1] < 0.0f ||
          baryCoords[2] < 0.0f)
          continue;
        float currz = zABC*pixv;
        V3 currCol;
        float currDen = msiDenABC*pixv;
        if (!texture) {
//          currCol = colABC*pixv;
          currCol = V3(redABC*pixv, greenABC*pixv, blueABC*pixv)/currDen;
        }
        else {
          float currs = texABCs[0]*pixv;
          float currt = texABCs[1]*pixv;
          currs = texABCsm[0]*pixv/currDen;
          currt = texABCsm[1]*pixv/currDen;
          currCol.SetFromUIntColor(texture->Lookup(currs, currt));
        }

        V3 currP = ppc->GetPoint1w(pixv[0], pixv[1], currz);
        V3 projP;
        if (projPPC->ProjectInFrame(currP, projP)) {
          currCol.SetFromUIntColor(projFB->LookupNN(projP[0], projP[1]));
        }
        fb->SetIfCloser(V3(pixv[0], pixv[1], currz), currCol);
      }
    }
  }

}

V3 TMesh::GetCenter() {

  V3 ret(0.0f, 0.0f, 0.0f);
  for (int i = 0; i < verticesN; i++) {
    ret = ret + vertices[i];
  }
  ret = ret / (float) verticesN;
  return ret;

}

void TMesh::Translate(V3 trans) {

  for (int i = 0; i < verticesN; i++) {
    vertices[i] = vertices[i] + trans;
  }

}

void TMesh::Scale(float scf) {

  for (int i = 0; i < verticesN; i++) {
    vertices[i] = vertices[i]*scf;
  }

}

BBox TMesh::ComputeBBox() {

  BBox ret;
  for (int i = 0; i < verticesN; i++)
    ret.AddPoint(vertices[i]);

  return ret;

}

void TMesh::PlaceAndScale(V3 newCenter, 
                          float bboxDiag) {

  V3 center = GetCenter();
  BBox bbox = ComputeBBox();
  Translate(center*-1.0f);
  Scale(bboxDiag/
    (bbox.corners[0]-bbox.corners[1]).Length());
  Translate(newCenter);

}

M3x3 TMesh::ComputeBarycentricInverse(V3 pvs[3]) {

  M3x3 ret;
  M3x3 vM;
  vM[0] = V3(pvs[0][0], pvs[1][0], pvs[2][0]);
  vM[1] = V3(pvs[0][1], pvs[1][1], pvs[2][1]);
  vM[2] = V3(1.0f, 1.0f, 1.0f);
  ret = vM.Invert();
  return ret;

}

M3x3 TMesh::ComputeSSI(V3 pvs[3]) {

  M3x3 ret;
  M3x3 vM;
  vM[0] = pvs[0]; vM[0][2] = 1.0f;
  vM[1] = pvs[1]; vM[1][2] = 1.0f;
  vM[2] = pvs[2]; vM[2][2] = 1.0f;
  ret = vM.Invert();
  return ret;

}

M3x3 TMesh::ComputeMMI(unsigned int inds[3], PPC *ppc) {

  M3x3 M0;
  M0.SetColumn(0, vertices[inds[0]]-ppc->C);
  M0.SetColumn(1, vertices[inds[1]]-ppc->C);
  M0.SetColumn(2, vertices[inds[2]]-ppc->C);
  M3x3 M1;
  M1.SetColumn(0, ppc->a);
  M1.SetColumn(1, ppc->b);
  M1.SetColumn(2, ppc->c);
  M3x3 ret = M0.Invert()*M1;
  return ret;

}

void TMesh::SetTCsQuad() {

  if (!tcs)
    tcs = new float[2*verticesN];

  tcs[0] = 0.0f;
  tcs[1] = 1.0f;
  tcs[2] = 1.0f;
  tcs[3] = 1.0f;
  tcs[4] = 1.0f;
  tcs[5] = 0.0f;
  tcs[6] = 0.0f;
  tcs[7] = 0.0f;


}

V3 TMesh::GetNumABC(M3x3 qM, V3 rv) {

  V3 ret(qM.GetColumn(0)*rv, qM.GetColumn(1)*rv, 
    qM.GetColumn(2)*rv);

  return ret;

}
